@@ -20,8 +20,8 @@ TAG_CHARS = [
     CHAR_TAG_END
 ]
 
-TAG_CLOSING = '/>'
-TAG_END = '</'
+TAG_CLOSING = CHAR_TAG_CLOSE + CHAR_TAG_END
+TAG_END = CHAR_TAG_OPEN + CHAR_TAG_CLOSE
 
 SPACE_CHARS = [
     0x20,
