@@ -11,11 +11,49 @@ def appendToken(context, type, text):
         })
 
 
+def isAlpha(char: str) -> bool:
+    code = ord(char)
+    return (code >= 97 and code <= 122) or (code >= 65 and code <= 90)
+
+
 def isDigit(char: str) -> bool:
     code = ord(char)
     return code >= 0x30 and code <= 0x39
 
 
+def isAlphaOrSlash(char: str) -> bool:
+    return isAlpha(char) or char == constants.CHAR_TAG_CLOSE
+
+
+def isAlphaNum(char: str) -> bool:
+    return isAlpha(char) or isDigit(char)
+
+
+def isElementNameChar(char: str) -> bool:
+    code = ord(char)
+    return (
+        code == 45 or # '-'
+        code == 46 or # '.'
+        (code >= 48 and code <= 57) or # 0..9
+        code == 95 or # '_'
+        (code >= 97 and code <= 122) or # a..z
+        (code >= 65 and code <= 90) or # A..Z
+        code == 0xB7 or
+        (code >= 0xc0 and code <= 0xD6) or
+        (code >= 0xD8 and code <= 0xF6) or
+        (code >= 0xF8 and code <= 0x37D) or
+        (code >= 0x37F and code <= 0x1FFF) or
+        (code >= 0x200C and code <= 0x200D) or
+        (code >= 0x203F and code <= 0x2040) or
+        (code >= 0x2070 and code <= 0x218F) or
+        (code >= 0x2C00 and code <= 0x2FEF) or
+        (code >= 0x3001 and code <= 0xD7FF) or
+        (code >= 0xF900 and code <= 0xFDCF) or
+        (code >= 0xFDF0 and code <= 0xFFFD) or
+        (code >= 0x10000 and code <= 0xEFFFF)
+    )
+
+
 def isSpace(char: str) -> bool:
     code = ord(char)
     return code in constants.SPACE_CHARS or \
@@ -39,6 +77,10 @@ def skipSpace(context, ret = False):
         appendToken(context, 'space', msg[start:context['i']])
 
 
+def recursion(context):
+    raise SyntaxError("Too much recursion at position {}".format(context['i']))
+
+
 def unexpected(char, index = None):
     if isinstance(char, dict):
         index = char['i']
@@ -60,8 +102,11 @@ class Parser:
         self.options = {
             'subnumeric_types': ['plural', 'selectordinal'],
             'submessage_types': ['plural', 'selectordinal', 'select'],
+            'maximum_depth': 50,
             'allow_tags': False,
+            'strict_tags': False,
             'tag_type': 'tag',
+            'include_indices': False,
             'loose_submessages': False,
             'allow_format_spaces': True,
             'require_other': True
@@ -78,7 +123,8 @@ class Parser:
         context = {
             'msg': input,
             'length': len(input),
-            'i': 0
+            'i': 0,
+            'depth': 0
         }
 
         if tokens is not None:
@@ -88,6 +134,13 @@ class Parser:
 
         try:
             return self._parseAST(context, None)
+        except RecursionError:
+            # Any RecursionError is also a syntax error
+            # because there is no reasonable reason to have
+            # such a deeply nested ICU MessageFormat string.
+            # Re-throw it as a SyntaxError as though we hit
+            # the maximum allowed depth.
+            raise recursion(context)
         except IndexError:
             # Any IndexErrors we may encounter are actually
             # syntax errors as we try to read parts of the
@@ -119,9 +172,7 @@ class Parser:
             if parent and self.options['allow_tags'] and msg[i:i+len(constants.TAG_END)] == constants.TAG_END:
                 break
 
-            result = self._parsePlaceholder(context, parent)
-            if result:
-                out.append(result)
+            out.append(self._parsePlaceholder(context, parent))
             start = context['i']
             text = self._parseText(context, parent)
             if text:
@@ -131,6 +182,34 @@ class Parser:
         return out
 
 
+    def _canReadTag(self, context, parent):
+        msg = context['msg']
+        length = context['length']
+        start = context['i']
+        current = context['i']
+        if not self.options['allow_tags']:
+            return False
+
+        char = msg[current] if current < length else None
+        if char != constants.CHAR_TAG_OPEN:
+            return False
+
+        if self.options['strict_tags']:
+            return True
+
+        current += 1
+        char = msg[current] if current < length else None
+
+        # We're trying to close a tag.
+        if char == constants.CHAR_TAG_CLOSE:
+            return True
+
+        if isAlpha(char):
+            return True
+
+        return False
+
+
     def _parseText(self, context, parent, is_arg_style = False):
         msg = context['msg']
         length = context['length']
@@ -148,7 +227,7 @@ class Parser:
 
             if char in constants.VAR_CHARS or \
                     (is_hash_special and char == constants.CHAR_HASH) or \
-                    (is_tag_special and char == constants.CHAR_TAG_OPEN) or \
+                    (is_tag_special and char == constants.CHAR_TAG_OPEN and self._canReadTag(context, parent)) or \
                     (is_arg_style and not allow_arg_spaces and is_space):
                 break
 
@@ -207,19 +286,28 @@ class Parser:
         return text
 
 
+    def _tokenIndices(self, token, start, end):
+        if self.options['include_indices']:
+            token['start'] = start
+            token['end'] = end
+        return token
+
+
     def _parsePlaceholder(self, context, parent):
         msg = context['msg']
         length = context['length']
         is_hash_special = parent and parent['type'] in self.options['subnumeric_types']
 
-        char = msg[context['i']] if context['i'] < length else None
+        start_idx = context['i']
+        char = msg[start_idx] if start_idx < length else None
         if is_hash_special and char == constants.CHAR_HASH:
             appendToken(context, 'hash', char)
             context['i'] += 1
-            return {
+            return self._tokenIndices({
                 'type': 'number',
-                'name': parent['name']
-            }
+                'name': parent['name'],
+                'hash': True
+            }, start_idx, context['i'])
 
         tag = self._parseTag(context, parent)
         if tag:
@@ -249,7 +337,7 @@ class Parser:
         if char == constants.CHAR_CLOSE:
             appendToken(context, 'syntax', char)
             context['i'] += 1
-            return token
+            return self._tokenIndices(token, start_idx, context['i'])
 
         if char != constants.CHAR_SEP:
             raise expected(SEP_OR_CLOSE, context)
@@ -274,7 +362,7 @@ class Parser:
                 raise expected('{} sub-messages'.format(ttype), context)
 
             context['i'] += 1
-            return token
+            return self._tokenIndices(token, start_idx, context['i'])
 
         if char != constants.CHAR_SEP:
             raise expected(SEP_OR_CLOSE, context)
@@ -326,7 +414,7 @@ class Parser:
 
         appendToken(context, 'syntax', char)
         context['i'] += 1
-        return token
+        return self._tokenIndices(token, start_idx, context['i'])
 
 
     def _parseTag(self, context, parent):
@@ -336,6 +424,7 @@ class Parser:
         msg = context['msg']
         length = context['length']
         i = context['i']
+        start_idx = i
         char = msg[i] if i < length else None
 
         if char != constants.CHAR_TAG_OPEN:
@@ -344,13 +433,18 @@ class Parser:
         if msg[i:i + len(constants.TAG_END)] == constants.TAG_END:
             raise unexpected(constants.TAG_END, i)
 
-        appendToken(context, 'syntax', char)
         context['i'] += 1
 
         name = self._parseName(context, True)
         if not name:
+            if not self.options['strict_tags']:
+                context['i'] = start_idx
+                return
+
             raise expected('tag name', context)
 
+        appendToken(context, 'syntax', char)
+
         token = {
             'type': self.options['tag_type'],
             'name': name
@@ -362,7 +456,7 @@ class Parser:
         if i < length and msg[i:i + len(constants.TAG_CLOSING)] == constants.TAG_CLOSING:
             appendToken(context, 'syntax', constants.TAG_CLOSING)
             context['i'] += len(constants.TAG_CLOSING)
-            return token
+            return self._tokenIndices(token, start_idx, context['i'])
 
         char = msg[i] if i < length else None
         if char != constants.CHAR_TAG_END:
@@ -395,8 +489,7 @@ class Parser:
 
         appendToken(context, 'syntax', char)
         context['i'] += 1
-
-        return token
+        return self._tokenIndices(token, start_idx, context['i'])
 
 
     def _parseName(self, context, is_tag = False):
@@ -446,6 +539,8 @@ class Parser:
         length = context['length']
         options = {}
 
+        context['depth'] += 1
+
         while context['i'] < length and msg[context['i']] != constants.CHAR_CLOSE:
             selector = self._parseName(context)
             if not selector:
@@ -456,6 +551,8 @@ class Parser:
             options[selector] = self._parseSubmessage(context, parent)
             skipSpace(context)
 
+        context['depth'] -= 1
+
         if not options:
             return None
 
@@ -477,6 +574,9 @@ class Parser:
 
 
     def _parseSubmessage(self, context, parent):
+        if context['depth'] >= self.options['maximum_depth']:
+            raise recursion(context)
+
         msg = context['msg']
         length = context['length']
         if context['i'] >= length or msg[context['i']] != constants.CHAR_OPEN:
