@@ -5,7 +5,9 @@ from pyicumessageformat import Parser
 ## Setup
 
 parser = Parser()
-tag_parser = Parser({'allow_tags': True})
+tag_strict_parser = Parser({'allow_tags': True, 'strict_tags': True, 'include_indices': True})
+tag_parser = Parser({'allow_tags': True, 'include_indices': True})
+idx_parser = Parser({'include_indices': True})
 
 def token(type, text):
     return {
@@ -48,7 +50,7 @@ tHash = token('hash', '#')
 def tokensToString(tokens):
     return ''.join(x['text'] for x in tokens)
 
-def parse(input, tokens = None):
+def parsePlain(input, tokens = None):
     out = parser.parse(input, tokens)
     if tokens:
         assert tokensToString(tokens) == input
@@ -60,6 +62,18 @@ def parseTags(input, tokens = None):
         assert tokensToString(tokens) == input
     return out
 
+def parseStrictTags(input, tokens = None):
+    out = tag_strict_parser.parse(input, tokens)
+    if tokens:
+        assert tokensToString(tokens) == input
+    return out
+
+def parse(input, tokens = None):
+    out = idx_parser.parse(input, tokens)
+    if tokens:
+        assert tokensToString(tokens) == input
+    return out
+
 
 ## The Tests
 
@@ -79,7 +93,32 @@ def test_input_types():
 
 def test_hello_name():
     tokens = []
-    assert parse('Hello, {name}!', tokens) == ['Hello, ', {'name': 'name'}, '!']
+    assert parsePlain('Hello, {name}!', tokens) == ['Hello, ', {'name': 'name'}, '!']
+    assert tokens == [
+        tText('Hello, '),
+        tOpen,
+            tName('name'),
+        tClose,
+        tText('!')
+    ]
+
+def test_hello_indices():
+    tokens = []
+    input = 'Hello, {name}!'
+    result = parse(input, tokens)
+    assert result == [
+        'Hello, ',
+        {
+            'name': 'name',
+            'start': 7,
+            'end': 13
+        },
+        '!'
+    ]
+
+    token = result[1]
+    assert input[token['start']: token['end']] == '{name}'
+
     assert tokens == [
         tText('Hello, '),
         tOpen,
@@ -90,7 +129,12 @@ def test_hello_name():
 
 def test_n_number():
     tokens = []
-    assert parse('{n, number}', tokens) == [{'name': 'n', 'type': 'number'}]
+    assert parse('{n, number}', tokens) == [{
+        'name': 'n',
+        'type': 'number',
+        'start': 0,
+        'end': 11
+    }]
     assert tokens == [
         tOpen,
             tName('n'),
@@ -105,7 +149,9 @@ def test_n_number_percent():
     assert parse('{num, number, percent }', tokens) == [{
         'name': 'num',
         'type': 'number',
-        'format': 'percent'
+        'format': 'percent',
+        'start': 0,
+        'end': 23
     }]
     assert tokens == [
         tOpen,
@@ -125,12 +171,20 @@ def test_plural_photos():
     assert parse('{numPhotos, plural, =0{no photos} =1{one photo} other{# photos}}', tokens) == [{
         'name': 'numPhotos',
         'type': 'plural',
+        'start': 0,
+        'end': 64,
         'offset': 0,
         'options': {
             '=0': ['no photos'],
             '=1': ['one photo'],
             'other': [
-                {'name': 'numPhotos', 'type': 'number'},
+                {
+                    'name': 'numPhotos',
+                    'type': 'number',
+                    'hash': True,
+                    'start': 54,
+                    'end': 55
+                },
                 ' photos'
             ]
         }
@@ -167,11 +221,19 @@ def test_plural_offset():
         'name': 'numGuests',
         'type': 'plural',
         'offset': 1,
+        'start': 0,
+        'end': 80,
         'options': {
             '=0': ['no party'],
             'one': ['host and a guest'],
             'other': [
-                {'name': 'numGuests', 'type': 'number'},
+                {
+                    'name': 'numGuests',
+                    'type': 'number',
+                    'hash': True,
+                    'start': 70,
+                    'end': 71
+                },
                 ' guests'
             ]
         }
@@ -210,6 +272,8 @@ def test_negative_offset():
     assert parse('{n, plural, offset:-12 other{x}}', tokens) == [{
         'name': 'n',
         'type': 'plural',
+        'start': 0,
+        'end': 32,
         'offset': -12,
         'options': {
             'other': ['x']
@@ -257,22 +321,48 @@ def test_ordinal():
     assert parse('{rank, selectordinal, one {#st} two {#nd} few {#rd} other {#th}}', tokens) == [{
         'name': 'rank',
         'type': 'selectordinal',
+        'start': 0,
+        'end': 64,
         'offset': 0,
         'options': {
             'one': [
-                {'name': 'rank', 'type': 'number'},
+                {
+                    'name': 'rank',
+                    'type': 'number',
+                    'hash': True,
+                    'start': 27,
+                    'end': 28
+                },
                 'st'
             ],
             'two': [
-                {'name': 'rank', 'type': 'number'},
+                {
+                    'name': 'rank',
+                    'type': 'number',
+                    'hash': True,
+                    'start': 37,
+                    'end': 38
+                },
                 'nd'
             ],
             'few': [
-                {'name': 'rank', 'type': 'number'},
+                {
+                    'name': 'rank',
+                    'type': 'number',
+                    'hash': True,
+                    'start': 47,
+                    'end': 48
+                },
                 'rd'
             ],
             'other': [
-                {'name': 'rank', 'type': 'number'},
+                {
+                    'name': 'rank',
+                    'type': 'number',
+                    'hash': True,
+                    'start': 59,
+                    'end': 60
+                },
                 'th'
             ]
         }
@@ -320,6 +410,8 @@ def test_select():
     assert parse('{gender, select, female {woman} male {man} other {person}}', tokens) == [{
         'name': 'gender',
         'type': 'select',
+        'start': 0,
+        'end': 58,
         'options': {
             'female': ['woman'],
             'male': ['man'],
@@ -359,6 +451,8 @@ def test_custom_one():
     assert parse('{a, custom, one}', tokens) == [{
         'name': 'a',
         'type': 'custom',
+        'start': 0,
+        'end': 16,
         'format': 'one'
     }]
     assert tokens == [
@@ -378,6 +472,8 @@ def test_weird_tags():
     assert parse('{<0/>,</>,void}', tokens) == [{
         'name': '<0/>',
         'type': '</>',
+        'start': 0,
+        'end': 15,
         'format': 'void'
     }]
     assert tokens == [
@@ -395,6 +491,8 @@ def test_weird_with_tags():
     assert parseTags('{<0/>,</>,void}', tokens) == [{
         'name': '<0/>',
         'type': '</>',
+        'start': 0,
+        'end': 15,
         'format': 'void'
     }]
     assert tokens == [
@@ -461,7 +559,8 @@ def test_no_loose():
 def test_loose_with_tags():
     x = Parser({
         'loose_submessages': True,
-        'allow_tags': True
+        'allow_tags': True,
+        'include_indices': True
     })
 
     input = '{a,<,>{click here}}'
@@ -469,6 +568,8 @@ def test_loose_with_tags():
     assert x.parse(input, tokens) == [{
         'name': 'a',
         'type': '<',
+        'start': 0,
+        'end': 19,
         'options': {
             '>': ['click here']
         }
@@ -498,8 +599,15 @@ def test_simple_xml():
     assert parseTags('<a><i/>here</a>', tokens) == [{
         'name': 'a',
         'type': 'tag',
+        'start': 0,
+        'end': 15,
         'contents': [
-            {'name': 'i', 'type': 'tag'},
+            {
+                'name': 'i',
+                'type': 'tag',
+                'start': 3,
+                'end': 7
+            },
             'here'
         ]
     }]
@@ -531,34 +639,70 @@ def test_escaping():
     assert parse('#', []) == ['#']
     assert parse("'", []) == ["'"]
     assert parse("{0} '{1}' {2}", []) == [
-        {'name': '0'},
+        {
+            'name': '0',
+            'start': 0,
+            'end': 3
+        },
         ' {1} ',
-        {'name': '2'}
+        {
+            'name': '2',
+            'start': 10,
+            'end': 13
+        }
     ]
     assert parse("{0} '{1} {2}", []) == [
-        {'name': '0'},
+        {
+            'name': '0',
+            'start': 0,
+            'end': 3
+        },
         ' {1} {2}'
     ]
     assert parse("{0} ''{1} {2}", []) == [
-        {'name': '0'},
+        {
+            'name': '0',
+            'start': 0,
+            'end': 3
+        },
         " '",
-        {'name': '1'},
+        {
+            'name': '1',
+            'start': 6,
+            'end': 9
+        },
         ' ',
-        {'name': '2'}
+        {
+            'name': '2',
+            'start': 10,
+            'end': 13
+        }
     ]
     assert parse("So, '{Mike''s Test}' is real.", []) == ["So, {Mike's Test} is real."]
     assert parse("You've done it now, {name}.", []) == [
         "You've done it now, ",
-        {'name': 'name'},
+        {
+            'name': 'name',
+            'start': 20,
+            'end': 26
+        },
         '.'
     ]
     assert parse("{n,plural,other{#'#'}}", []) == [{
         'name': 'n',
         'type': 'plural',
+        'start': 0,
+        'end': 22,
         'offset': 0,
         'options': {
             'other': [
-                {'name': 'n', 'type': 'number'},
+                {
+                    'name': 'n',
+                    'type': 'number',
+                    'hash': True,
+                    'start': 16,
+                    'end': 17
+                },
                 '#'
             ]
         }
@@ -569,6 +713,8 @@ def test_escape_format():
     assert parse("{n,date,'a style'}", tokens) == [{
         'name': 'n',
         'type': 'date',
+        'start': 0,
+        'end': 18,
         'format': 'a style'
     }]
     assert tokens == [
@@ -588,10 +734,14 @@ def test_mixed_tags_placeholders():
         {
             'name': 'boldThis',
             'type': 'tag',
+            'start': 13,
+            'end': 84,
             'contents': [
                 {
                     'name': 'price',
                     'type': 'number',
+                    'start': 23,
+                    'end': 73,
                     'format': '::currency/USD precision-integer'
                 }
             ]
@@ -600,10 +750,14 @@ def test_mixed_tags_placeholders():
         {
             'name': 'link',
             'type': 'tag',
+            'start': 90,
+            'end': 136,
             'contents': [
                 {
                     'name': 'pct',
                     'type': 'number',
+                    'start': 96,
+                    'end': 120,
                     'format': '::percent'
                 },
                 ' discount'
@@ -651,6 +805,9 @@ def test_throws_extra_close():
     with pytest.raises(SyntaxError, match='Unexpected "}"'):
         parse('}')
 
+    with pytest.raises(SyntaxError, match='Unexpected "}"'):
+        parse('{test}}')
+
 def test_throws_no_close():
     with pytest.raises(SyntaxError, match='Expected , or }'):
         parse('{a')
@@ -735,7 +892,9 @@ def test_throws_missing_closing_brace():
 
 def test_throws_missing_tag_id():
     with pytest.raises(SyntaxError, match='Expected tag name'):
-        parseTags('<>')
+        parseStrictTags('<>')
+
+    parseTags('<>')
 
 def test_throws_missing_tag_end():
     with pytest.raises(SyntaxError, match='Expected > or />'):
@@ -757,6 +916,20 @@ def test_throws_mismatch_tag():
     with pytest.raises(SyntaxError, match='Expected </a>'):
         parseTags('<a></b>')
 
+def test_loose_tags():
+    with pytest.raises(SyntaxError, match='Expected tag name'):
+        parseStrictTags('< {test}')
+
+    parseTags('< {test}')
+
+def test_closing_tags():
+    with pytest.raises(SyntaxError, match='Unexpected "</"'):
+        parseStrictTags('</3')
+
+    with pytest.raises(SyntaxError, match='Unexpected "</"'):
+        parseTags('</3')
+
+
 def test_custom_require_other():
     x = Parser({
         'require_other': ['plural']
@@ -782,3 +955,30 @@ def test_custom_require_other():
 
     assert x.parse('{n,select,cake{lie}}')
     assert x.parse('{n,plural,one{two}}')
+
+def test_recursion():
+    start = '{a,b,c{{a}'
+    end = '}}'
+
+    x = Parser({
+        'loose_submessages': True,
+        'require_other': False
+    })
+
+    inp = ''
+    out = ''
+    for i in range(1000):
+        inp = inp + start
+        out = out + end
+
+    with pytest.raises(SyntaxError, match='Too much recursion'):
+        x.parse(inp + out)
+
+    x = Parser({
+        'loose_submessages': True,
+        'require_other': False,
+        'maximum_depth': 99999
+    })
+
+    with pytest.raises(SyntaxError, match='Too much recursion'):
+        x.parse(inp + out)
